import { ParsedRouterRequest, ParsedSocketRequest, UnparsedRouterResponse, UnparsedSocketResponse } from '../types';

import { status } from '@grpc/grpc-js';
import { Status } from '@grpc/grpc-js/build/src/constants';

export type RouterRequestHandler = (call: ParsedRouterRequest) => Promise<UnparsedRouterResponse>;
export type SocketRequestHandler = (call: ParsedSocketRequest) => Promise<UnparsedSocketResponse>;
export type RequestHandlers = RouterRequestHandler | SocketRequestHandler;

export function wrapCallObjectForRouter(call: any): ParsedRouterRequest {
  return {
    request: {
      params: call.request,
      path: '',
      headers: {},
      context: {},
    },
  };
}

function generateLog(routerRequest: boolean, requestReceive: number, call: any, status?: Status) {
  console.log(
    routerRequest ? 'Request: ' : 'Socket: ' +
    routerRequest ? call.request.path : call.request.event +
      (routerRequest ? 'socket: ' + call.request.socket : '') +
      ' ' +
      status ?? '200' +
      ' ' +
      (Date.now() - requestReceive),
  );
}

function parseRequestData(data: any) {
  if (typeof data === 'string' && data.length !== 0) {
    return JSON.parse(data);
  } else if (typeof data === 'string') {
    return {};
  }
}

export function wrapRouterGrpcFunction(
  fun: RequestHandlers,
): (call: any, callback: any) => void {
  return (call: any, callback: any) => {
    let requestReceive = Date.now();
    let routerRequest = true;
    try {
      call.request.context = parseRequestData(call.request.context);
      call.request.params = parseRequestData(call.request.params);

      routerRequest = !call.request.hasOwnProperty('event');
      if (routerRequest) {
        call.request.headers = parseRequestData(call.request.headers);
      }
    } catch (e) {
      generateLog(routerRequest, requestReceive, call, status.INTERNAL);
      console.log(e.message ?? 'Something went wrong');
      return callback({
        code: status.INTERNAL,
        message: e.message ?? 'Something went wrong',
      });
    }

    fun(call)
      .then((r) => {
        if (!r) return;
        if (routerRequest) {
          if (typeof r === 'string') {
            callback(null, { result: r });
          } else {
            if (r.result || r.redirect) {
              callback(null, {
                redirect: r.redirect ?? undefined,
                result: r.result ? JSON.stringify(r.result) : undefined,
              });
            } else {
              callback(null, { result: JSON.stringify(r) });
            }
          }
        } else {
          if (r.hasOwnProperty('data')) (r as any).data = JSON.stringify((r as any).data);
          callback(null, r);
        }
        generateLog(routerRequest, requestReceive, call, undefined);
      })
      .catch((error) => {
        generateLog(routerRequest, requestReceive, call, error.code ?? status.INTERNAL);
        console.log(error.message ?? 'Something went wrong');
        callback({
          code: error.code ?? status.INTERNAL,
          message: error.message ?? 'Something went wrong',
        });
      });
  };
}
